#!/bin/python3

'''Provides the `Compiler` class'''

#> Imports
import re
import codecs
import struct
import typing
import buffer_matcher
from codecs import escape_decode
from pathlib import Path
from functools import singledispatchmethod

from . import nodes
from . import util
from .basic_compiler import RE_FLAGS
#</Imports

#> Header >/
__all__ = ('Compiler',)

class Compiler:
    '''Compiles grammars'''
    __slots__ = ('grammar', 'base_path')

    grammar: dict[bytes, nodes.Node]
    base_path: Path

    def __init__(self, grammar: dict[bytes, nodes.Node],
                 base_path: Path = Path(__file__).parent):
        self.grammar = grammar
        self.base_path = base_path

    # Compiling
    @singledispatchmethod
    def compile(self, src: Path | typing.BinaryIO | bytes | buffer_matcher.SimpleBufferMatcher | dict, *,
                source: Path | str | None = None, bind: bool = True) -> dict[bytes, nodes.Node]:
        '''Compiles grammars from various sources'''
        raise TypeError(f'Cannot compile from type {type(src).__qualname__}')

    ## Intermediate source types
    @compile.register
    def compile_path(self, path: Path, *, source: Path | str | None = None, **kwargs) -> dict[bytes, nodes.Node]:
        '''
            Compiles from a path
            See `.compile()` or `.compile_buffermatcher()` for `kwargs`
        '''
        return self.compile_bytes(path.read_bytes(), source=path if source is None else source, **kwargs)
    @compile.register
    def compile_file(self, file: typing.BinaryIO, *, source: Path | str | None = None, **kwargs) -> dict[bytes, nodes.Node]:
        '''
            Compiles from a file
            See `.compile()` or `.compile_buffermatcher()` for `kwargs`
        '''
        return self.compile_bytes(file.read(), source=file.name if source is None else source, **kwargs)
    @compile.register
    def compile_bytes(self, data: bytes, *, source: Path | str | None = None, **kwargs) -> dict[bytes, nodes.Node]:
        '''
            Compiles from bytes
            See `.compile()` or `.compile_buffermatcher()` for `kwargs`
        '''
        return self.compile_buffermatcher(buffer_matcher.DynamicBufferMatcher(data),
                                          source='<bytes>' if source is None else source, **kwargs)
    @compile.register
    def compile_buffermatcher(self, bm: buffer_matcher.SimpleBufferMatcher,
                              source: Path | str | None = None, bind: bool = True,
                              precompile_only: bool = False) -> dict[bytes, nodes.Node]:
        '''Compiles a grammar (dict of names and nodes) from a buffer matcher'''
        if source is None: source = '<buffermatcher>'
        try: pre = self.pre_process(bm, source=source)
        except Exception as e:
            e.add_note(f'Whilst pre-processing grammar from: {source}')
            raise e
        if precompile_only: return pre
        try: return self.post_process_compile(pre, bind=bind)
        except Exception as e:
            e.add_note(f'Whilst post-process compiling grammar from: {source}')
            raise e

    ## Pre-process
    def pre_process(self, bm: buffer_matcher.SimpleBufferMatcher,
                    source: Path | str | None = None) -> dict[bytes, dict]:
        '''
            Reads a buffer matcher and parses it into a format suitable for compilation
            Also handles pragmas
        '''
        working = {}
        while True:
            # Discard junk characters and check for EOF
            while bm.match(util.WHITESPACE_PATT) or self.grammar[b'COMMENT'](bm):
                pass # ignore whitespace and comments
            if not bm.peek(1): break # EOF
            # Handle pragmas
            if p := self.grammar[b'PRAGMA'](bm):
                try: self.handle_pragma(p['type'], p['args'], working=working, bm=bm, source=source)
                except Exception as e:
                    e.add_note(f'In pragma at {bm.pos} ({bm.lno}:{bm.cno})')
                    raise e
                continue
            # Parse statements
            stmt = self.grammar[b'STATEMENT'](bm, stealer=True)
            working[stmt['name']] = stmt['expr']
        return working
    ### Pragma
    def handle_pragma(self, type_: bytes, args: bytes, *, working: dict[bytes, dict],
                      bm: buffer_matcher.SimpleBufferMatcher, source: Path | str | None = None) -> None:
        '''Handles a pragma statement during compilation'''
        match type_:
            case b'include':
                if not args:
                    raise SyntaxError('Include pragma requires args')
                f = Path(args.decode())
                if f.is_absolute():
                    working.update(self.compile_path(f, bind=False, precompile_only=True))
                    return
                if isinstance(source, Path) and (nf := (source.parent / f)).is_file(): pass
                elif (nf := (Path.cwd() / f)).is_file(): pass
                elif (nf := (self.base_path / f)).is_file(): pass
                else: raise FileNotFoundError(f'Cannot resolve relative file from include: {f}')
                working.update(self.compile_path(nf, bind=False, precompile_only=True))
            case _:
                raise TypeError(f'Unknown pragma type {type_!r}')

    ## Compile
    def post_process_compile(self, working: dict[bytes, dict], *, bind: bool = True) -> dict[bytes, nodes.Node]:
        '''Compiles pre-processed grammar into nodes'''
        grammar = {name: self.compile_node(b'group', node)
                   for name,node in working.items()}
        if bind: util.bind_nodes(grammar)
        return grammar

    def compile_expr(self, name: bytes, expr: dict) -> nodes.Node:
        '''Compiles a node's expression and names it'''
        node = self.compile_node(**expr)
        node.name = None if name is None else name.decode()
        return node

    def compile_node(self, type: bytes, val: dict | bytes | None = None) -> nodes.Node:
        '''Compiles a node's expression'''
        match type:
            case b'group' | b'group_ws_sensitive' | b'union':
                subnodes = tuple(self.compile_expr(**expr) for expr in val)
                if type == b'union':
                    return nodes.NodeUnion(*subnodes)
                return nodes.NodeGroup(*subnodes, keep_whitespace=(type == b'group_ws_sensitive'))
            case b'range' | b'range_ws_sensitive':
                return nodes.NodeRange(node=self.compile_expr(**val['node']),
                    min=int(val['min'] or 0), max=(None if val['max'] is None else int(val['max'])),
                    keep_whitespace=(type == b'range_ws_sensitive'))
            case b'string':
                return nodes.StringNode(escape_decode(val)[0])
            case b'pattern':
                flags = re.NOFLAG
                for f in struct.unpack(f'{len(val["flags"])}c', val['flags']): # iterate through bytes
                    flag = RE_FLAGS.get(f, None)
                    if flag is None:
                        raise ValueError(f'Unknown regular expression flag {f!r}')
                    flags |= flag
                return nodes.PatternNode(re.compile(val['pattern'], flags),
                                         None if val['group'] is None else int(val['group']))
            case b'stealer':
                return nodes.Stealer()
            case b'context':
                return nodes.Context(escape_decode(val['str'])[0] if 'str' in val else val['raw'])
            case b'noderef':
                return nodes.NodeRef(val)
            case _:
                raise TypeError(f'Unknown node type {type!r}')



#!/bin/python3

'''Caustic's lexing/grammar framework'''

#> Imports
from pathlib import Path
#</Imports

#> Package >/
__all__ = ('basic_compiler', 'nodes', 'serialize', 'util',
           'Compiler', 'saved_compiler', 'load_compiler', 'save_compiler')

from . import basic_compiler
from . import nodes
from . import serialize
from . import util

from .compiler import Compiler

saved_compiler: Compiler | None = None
def load_compiler(type_: type = Compiler, base_path: Path = Path(__file__).parent, *,
                  file_cache: bool = True) -> Compiler:
    '''
        Loads the advanced grammar compiler using `basic_compiler`,
            and sets `saved_compiler` to it
        If it was already loaded as `saved_compiler`, returns that object
        If `file_cache`, then checks for the file cache as well
    '''
    global saved_compiler
    if saved_compiler is not None: return saved_compiler
    if file_cache:
        pf = base_path/'precompiled_nodes.pkl'
        if pf.is_file():
            saved_compiler = Compiler(serialize.deserialize(pf.read_bytes()), base_path)
            return saved_compiler
    gf = (Path(__file__).parent / 'grammar.cag')
    if not gf.is_file():
        raise FileNotFoundError(f'{gf} does not exist as a file')
    from buffer_matcher import DynamicBufferMatcher #
    saved_compiler = type_(basic_compiler.compile(
        DynamicBufferMatcher(gf.read_bytes())), base_path)
    return saved_compiler

def save_compiler(comp: Compiler, base_path: Path = Path(__file__).parent) -> None:
    '''Saves the compiler to a persistent file'''
    (base_path/'precompiled_nodes.pkl').write_bytes(serialize.serialize(comp.grammar))



#!/bin/python3

'''Provides nodes for matching grammar'''

#> Imports
import io
import re
import typing
from abc import ABCMeta, abstractmethod
from buffer_matcher import SimpleBufferMatcher
from collections import abc as cabc

from .util import WHITESPACE_PATT
from .util import NO_MATCH
#</Imports

#> Header >/
__all__ = ('NodeSyntaxError',
           'Node', 'NodeGroup', 'NodeUnion', 'NodeRange',
           'StringNode', 'PatternNode',
           'Stealer', 'Context', 'NodeRef')

# Exceptions
class NodeSyntaxError(SyntaxError):
    '''For when nodes fail to match something that must be matched'''
    __slots__ = ('node',)

    node: 'Node'
    bm: SimpleBufferMatcher

    def __init__(self, node: 'Node', bm: SimpleBufferMatcher, message: str):
        super().__init__(message)
        self.node = node; self.bm = bm
    def __str__(self) -> str:
        chain = self
        depth = 0
        with io.StringIO() as sio:
            sio.write('Node syntax exception chain (most recent failure last):')
            while chain is not None:
                sio.write(f'\n<{depth}>Node: {chain.node} failed @ {chain.bm.pos} ({chain.bm.lno+1}:{chain.bm.cno})\n')
                sio.write('\n'.join(chain.args))
                for n in getattr(chain, '__notes__', ()):
                    sio.write(f'\nNote: {n}')
                chain = chain.__cause__
                depth += 1
            return sio.getvalue()
# Nodes
## Base
class Node(metaclass=ABCMeta):
    '''The base class for all nodes'''
    __slots__ = ('name',)

    name: str | None

    def __init__(self, *, name: str | None = None):
        self.name = name

    def bind(self, nodes: dict[bytes, typing.Self]) -> None:
        '''Binds all sub-nodes, if possible'''
        if not hasattr(self, 'nodes'): return
        for node in self.nodes: node.bind(nodes)
    def unbind(self) -> None:
        '''Unbinds all sub-nodes'''
        if not hasattr(self, 'nodes'): return
        for node in self.nodes: node.unbind()

    @abstractmethod
    def __call__(self, bm: SimpleBufferMatcher, *, stealer: bool = False) -> object | dict[str, typing.Any]:
        '''Executes this node on `data`'''

    @abstractmethod
    def __str__(self) -> str: pass
    @abstractmethod
    def __repr__(self) -> str: pass

## Groups
class NodeGroup(Node):
    '''
        A group of nodes
        Discards whitespace between nodes if `keep_whitespace` is false
    '''
    __slots__ = ('nodes', 'keep_whitespace')

    nodes: tuple[Node, ...]
    keep_whitespace: bool

    def __init__(self, *nodes: Node, keep_whitespace: bool = False, **kwargs):
        super().__init__(**kwargs)
        self.nodes = nodes
        self.keep_whitespace = keep_whitespace

    def __call__(self, bm: SimpleBufferMatcher, stealer: bool = False) -> object | dict[str, typing.Any] | list[typing.Any] | None:
        save = bm.save_pos()
        results = []
        single_result = False
        after = None
        for i,n in enumerate(self.nodes):
            if not self.keep_whitespace:
                bm.match(WHITESPACE_PATT)
            if isinstance(n, Stealer):
                if not i:
                    se = SyntaxError('Cannot have a stealer at the beginning of a group')
                    se.add_note(str(self))
                    raise se
                stealer = True
                after = self.nodes[-1]
                continue
            # Execute node
            try: res = n(bm, stealer=stealer)
            except NodeSyntaxError as nse:
                raise NodeSyntaxError(self, bm, f'Node {i} failed underneath node-group') from nse
            if res is NO_MATCH:
                assert not stealer
                bm.load_pos(save)
                return NO_MATCH
            # Check how we should return results
            if n.name is None: # not assigned a name ("[name]:<node>")
                if isinstance(results, dict): continue # don't add it
                if not single_result: results.append(res)
            elif n.name: # name is not blank ("<name>:<node>")
                if isinstance(results, dict):
                    results[n.name] = res
                elif single_result:
                    te = TypeError(f'Conflicting return types: named result {n.name} cannot be added to single result')
                    te.add_note(str(n))
                    te.add_note(f'In {self}')
                    raise te
                else:
                    results = {n.name: res}
            else: # name is blank (":<node>")
                if isinstance(results, dict):
                    te = TypeError('Conflicting return types: single result cannot be added to named results')
                    te.add_note(str(n))
                    te.add_note(f'In {self}')
                    raise te
                single_result = True
                results = res
        if not results: return None
        return results

    def __str__(self) -> str:
        return f'{"" if self.name is None else f"{self.name}:"}{"({"[self.keep_whitespace]} {" ".join(map(str, self.nodes))} {")}"[self.keep_whitespace]}'
    def __repr__(self) -> str:
        return f'<{type(self).__qualname__} {self.name!r}{" [keep_whitespace]" if self.keep_whitespace else ""} {self.nodes!r}>'

class NodeUnion(Node):
    '''Matches any of its nodes'''
    __slots__ = ('nodes',)

    nodes: tuple[Node, ...]

    def __init__(self, *nodes: Node, **kwargs):
        super().__init__(**kwargs)
        self.nodes = nodes

    def __call__(self, bm: SimpleBufferMatcher, *, stealer: bool = False) -> object | dict[str, typing.Any]:
        for n in self.nodes:
            if (res := n(bm)) is not NO_MATCH:
                return res
        if stealer: raise NodeSyntaxError(self, bm, f'Expected union {self}')
        return NO_MATCH

    def __str__(self) -> str:
        return f'{"" if self.name is None else f"{self.name}:"}[ {" ".join(map(str, self.nodes))} ]'
    def __repr__(self) -> str:
        return f'<{type(self).__qualname__} {self.name!r} {self.nodes!r}>'

    @classmethod
    def from_kwargs(cls, kwargs: dict) -> typing.Self:
        return cls(*kwargs['nodes'], name=kwargs['name'])

class NodeRange(Node):
    '''
        Matches between `min` and `max` nodes,
            or any amount over `min` if `max` is `None`
    '''
    __slots__ = ('min', 'max', 'node', 'keep_whitespace')

    def __init__(self, node: Node, min: int | None, max: int | None, *, keep_whitespace: bool = False, **kwargs):
        super().__init__(**kwargs)
        if min is None: min = 0
        else: assert min >= 0, 'min should not be negative'
        assert (max is None) or (max >= 0), 'max should be None or more than or equal to min'
        self.min = min
        self.max = max
        self.node = node
        self.keep_whitespace = keep_whitespace

    def bind(self, nodes: dict[bytes, Node]) -> bool | None:
        '''Binds the underlying node, if applicable'''
        return self.node.bind(nodes)

    def __call__(self, bm: SimpleBufferMatcher, *, stealer: bool = False) -> object | list[typing.Any]:
        results = []
        save = bm.save_pos()
        for _ in range(self.min):
            try: results.append(self.node(bm, stealer=stealer))
            except NodeSyntaxError as nse:
                raise NodeSyntaxError(self, bm, f'Expected at least {self.min} of {self.node}') from nse
            if results[-1] is NO_MATCH:
                self.load_pos()
                return NO_MATCH
            if not self.keep_whitespace:
                bm.match(WHITESPACE_PATT)
        if self.max is None:
            while (res := self.node(bm)) is not NO_MATCH:
                results.append(res)
                if not self.keep_whitespace:
                    bm.match(WHITESPACE_PATT)
        else:
            for _ in range(self.min, self.max):
                res = self.node(bm)
                if res is NO_MATCH: break
                results.append(res)
                if not self.keep_whitespace:
                    bm.match(WHITESPACE_PATT)
        return results

    def __str__(self) -> str:
        return f'{self.min or ""}-{"" if self.max is None else {self.max}} {self.node}'
    def __repr__(self) -> str:
        return f'<{type(self).__qualname__} {self.name!r} {" [keep_whitespace]" if self.keep_whitespace else ""}{self.min!r} - {self.max!r}>'

## Real
class StringNode(Node):
    '''Matches a specific string'''
    __slots__ = ('string',)

    string: bytes

    def __init__(self, string: bytes, **kwargs):
        super().__init__(**kwargs)
        self.string = string
        if not self.string:
            raise ValueError('Cannot use an empty string')

    def __call__(self, bm: SimpleBufferMatcher, *, stealer: bool = False) -> object | bytes:
        if bm.match(self.string):
            return self.string
        if stealer:
            raise NodeSyntaxError(self, bm, f'Expected string {self}')
        return NO_MATCH

    def __str__(self) -> str:
        return f'"{"" if self.name is None else f"{self.name}:"}{self.string.decode(errors="backslashreplace").replace("\"", "\\\"")}"'
    def __repr__(self) -> str:
        return f'<{type(self).__qualname__} {self.name!r} {self.string!r}>'
class PatternNode(Node):
    '''Matches a pattern (regular expression)'''
    __slots__ = ('pattern', 'group')

    group: int | None
    pattern: re.Pattern

    def __init__(self, pattern: re.Pattern, group: int | None = None, **kwargs):
        super().__init__(**kwargs)
        self.pattern = pattern
        self.group = group

    def __call__(self, bm: SimpleBufferMatcher, *, stealer: bool = False) -> object | re.Match | bytes:
        if (m := bm.match(self.pattern)) is not None:
            return m.group(self.group) if self.group is not None else m
        if stealer:
            raise NodeSyntaxError(self, bm, f'Expected pattern {self}')
        return NO_MATCH

    FLAGS = {'i': re.IGNORECASE, 'm': re.MULTILINE, 's': re.DOTALL}
    def __str__(self) -> str:
        return (f'{"" if self.name is None else f"{self.name}:"}'
                f'{"" if self.group is None else self.group}/'
                f'{self.pattern.pattern.decode(errors="backslashreplace").replace("/", "\\/")}/'
                f'{"".join(f for f,v in self.FLAGS.items() if v & self.pattern.flags)}')
    def __repr__(self) -> str:
        return f'<{type(self).__qualname__} {self.pattern!r}{"" if self.group is None else f"[{self.group}]"}>'

## Meta
class Stealer(Node):
    '''Marks a special "stealer" node'''
    __slots__ = ()

    def __call__(self, *args, **kwargs):
        raise TypeError(f'Stealer nodes should not be called')

    def __str__(self) -> str: return '!'
    def __repr__(self) -> str: return f'<{type(self).__qualname__}>'
class Context(Node):
    '''Marks a special "context" node that always matches'''
    __slots__ = ('val',)

    val: typing.Any

    def __init__(self, val: typing.Any, **kwargs):
        assert val is not NO_MATCH, 'Cannot use NO_MATCH marker object for Context val'
        super().__init__(**kwargs)
        self.val = val

    def __call__(self, bm: SimpleBufferMatcher, *, stealer: bool = False) -> typing.Any:
        return self.val

    def __str__(self) -> str:
        return f'{"" if self.name is None else f"{self.name}:"}< {self.val} >'
    def __repr__(self) -> str:
        return f'<{type(self).__qualname__}>'
class NodeRef(Node):
    '''Marks a special "reference" node that "includes" another node'''
    __slots__ = ('target_name', 'target')

    target_name: bytes
    target: Node | None

    def __init__(self, target: bytes, **kwargs):
        super().__init__(**kwargs)
        self.target_name = target
        self.target = None

    @property
    def bound(self) -> bool: return self.target is not None
    def bind(self, targets: dict[bytes, Node]) -> bool:
        '''
            Attempts to bind this node to nodes in a dict
            If `.target_name` is not found in the dict, `False` is returned,
                otherwise `.target` is set to that node and `True` is returned
            Note: if this node was previously bound, that binding is removed,
                even if rebinding fails
        '''
        self.target = targets.get(self.target_name)
        return self.target is not None
    def unbind(self) -> None:
        '''Unbinds this node'''
        self.target = None

    def __call__(self, bm: SimpleBufferMatcher, *, stealer: bool = False) -> typing.Any:
        if not self.bound:
            raise TypeError(f'Cannot call an unbound NodeRef (node target {self.target_name} was never bound)')
        try: return self.target(bm, stealer)
        except NodeSyntaxError as nse:
            nse.add_note(f'Under reference {self}')
            raise nse

    def __str__(self) -> str:
        return f'@{self.target_name!r}'
    def __repr__(self) -> str:
        return (f'<{type(self).__qualname__} {self.name!r} target:{self.target_name!r} '
                f'{"[unbound]" if self.target is None else repr(self.target)}>')
